--  Chess piece drawing implementation based on OpenGL, GLFW, and 
--  the SOE library by Paul Liu and Duncan Coutts, using fragments
--  of SOE's accompanying drawing modules.
--
--  Author: Willem Heijltjes
-- 
--
--  This library is free software; you can redistribute it and/or
--  modify it under the terms of the GNU Lesser General Public
--  License as published by the Free Software Foundation; either
--  version 2.1 of the License, or (at your option) any later version.
--
--  This library is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
--  Lesser General Public License for more details.


module Picture_Parts (Picture,
                    display,
                    height, width,
                    flipH, flipV, rotateL, rotateR,
                    invert,
                    beside, above, superimpose,
                    repeatH, repeatV,
                    whiteSquare, greySquare, blackSquare, clearSquare,
                    bishop, king, knight, pawn, queen, rook, lambda, luke
                   ) where

import qualified Graphics.Rendering.OpenGL as GL
import Pictures
import Data.List as List
import Data.Set as Set
import Control.Monad( liftM, liftM2, liftM3 )
import Test.QuickCheck( Arbitrary (arbitrary, coarbitrary), elements )


-- Color regions

data CRegion = BLU -- bluescreen (background)
             | WSQ -- white square (background)
             | BSQ -- black square (background)
             | WPC -- white piece (foreground)
             | BPC -- black piece (foreground)
             | BOR -- piece borders (foreground)
             | RED -- bluescreen (background)
             | PRP -- bluescreen (background)
             | GRN -- bluescreen (background)
             | DGRN -- bluescreen (background)
             | LBLU -- bluescreen (background)
             | MGRN -- bluescreen (background)
             | MRED -- bluescreen (background)
             | LGRN -- bluescreen (background)
             | LRED -- bluescreen (background)

             | SKYA -- bluescreen (background)
             | SKYB -- bluescreen (background)
             | SKYC -- bluescreen (background)
             | SKYD -- bluescreen (background)

             | FLRA -- bluescreen (background)
             | FLRB -- bluescreen (background)
               deriving (Eq, Show, Ord)

gray :: Float -> GL.Color3 Float
gray a = GL.Color3 a a a

regionToRGB :: CRegion -> GL.Color3 Float
regionToRGB BLU = GL.Color3 0.2 0.6 0.8
regionToRGB WSQ = gray 1
regionToRGB BSQ = gray 0.5
regionToRGB WPC = gray 1
regionToRGB BPC = gray 0
regionToRGB BOR = gray 0.2

regionToRGB RED = GL.Color3 1 0.4 0.4
regionToRGB PRP = GL.Color3 0.6 0.0 0.0

regionToRGB GRN = GL.Color3 0.4 1 0.4
regionToRGB DGRN = GL.Color3 0.0 0.6 0.0

regionToRGB LBLU = GL.Color3 0.4 0.8 1

regionToRGB MGRN = GL.Color3 0.2 0.8 0.2
regionToRGB MRED = GL.Color3 0.8 0.2 0.2

regionToRGB LGRN = GL.Color3 0.8 1 0.8
regionToRGB LRED = GL.Color3 1 0.8 0.8

regionToRGB SKYA = GL.Color3 0.4 0.5 0.7
regionToRGB SKYB = GL.Color3 0.5 0.6 0.8
regionToRGB SKYC = GL.Color3 0.6 0.7 0.9
regionToRGB SKYD = GL.Color3 0.7 0.8 1

regionToRGB FLRA = GL.Color3 1 0.9 0.6
regionToRGB FLRB = GL.Color3 1 1 0.8


-- Colored shapes

data CShape = Sq CRegion Point Point -- Square
            | Pt CRegion Point       -- Point
              deriving (Show)

instance Eq CShape where
    Sq ca p1a p2a == Sq cb p1b p2b = rectNormal p1a p2a == rectNormal p1b p2b
    Pt ca psa == Pt cb psb = ca == cb && psa == psb

instance Ord CShape where
    compare (Sq ca p1a p2a) (Sq cb p1b p2b) 
            | rectNormal p1a p2a < rectNormal p1b p2b = LT
            | rectNormal p1a p2a > rectNormal p1b p2b = GT
            | otherwise = compare ca cb
    compare (Pt ca pa) (Pt cb pb)
            | pa < pb = LT
            | pa > pb = GT
            | otherwise = compare ca cb
    compare (Pt _ _) (Sq _ _ _) = LT
    compare (Sq _ _ _) (Pt _ _) = GT

rectNormal (x1,y1) (x2,y2) = ((min x1 x2, min y1 y2),(max x1 x2, max y1 y2))


-- Generators for QuickCheck

instance (Arbitrary a, Ord a) => Arbitrary (Set a) where
  arbitrary     = liftM fromList arbitrary
  coarbitrary n = undefined

instance Arbitrary CShape where
  arbitrary     = do x <- elements [True,False]
                     if x then liftM3 Sq arbitrary arbitrary arbitrary
                          else liftM2 Pt arbitrary arbitrary
  coarbitrary n = undefined


instance Arbitrary CRegion where
  arbitrary     = elements [BLU,WSQ,BSQ,WPC,BPC,BOR]
  coarbitrary n = undefined


-- Pictures

type Picture = (Point, Set CShape)

height, width :: Picture -> Int
height (p,_) = snd p
width  (p,_) = fst p

-- Main drawing function

display :: Picture -> IO ()
display ((x,y), pcs) = do w <- openWindow "Recursive Fish - By Luke Brown" (x,y)
                          drawInWindow w $ toGraphic $ shapeToDrawing BSQ (Sq BLU (0,0) (x,y))
                          drawInWindow w $ toGraphic $ foldl (>>=) (return BLU) (List.map (flip shapeToDrawing) $ sortByColor $ Set.toList pcs)
                          spaceClose w

toGraphic :: IO CRegion -> Graphic
toGraphic cr = Graphic (cr >> return ())

shapeToDrawing :: CRegion -> CShape -> IO CRegion
shapeToDrawing c1 (Sq c2 p1 p2) = (if c1 /= c2 then GL.color (regionToRGB c2) else return ()) >> 
                                  GL.renderPrimitive GL.Polygon (mapM_ pointToVertex (f p1 p2)) >>
                                  return c2
                                      where 
                                        f (x1,y1) (x2,y2) = [(x1,y1),(x2,y1),(x2,y2),(x1,y2)]
shapeToDrawing c1 (Pt c2 pt)    =  (if c1 /= c2 then GL.color (regionToRGB c2) else return ()) >>
                                   GL.renderPrimitive GL.Points (pointToVertex pt) >>
                                   return c2
                                      
pointToVertex :: Point -> IO()
pointToVertex (x,y) = GL.vertex (vertex3 (fromIntegral x) (fromIntegral y) 0)

sortByColor :: [CShape] -> [CShape]
sortByColor pcs = List.filter hasBGColor pcs ++ List.filter hasFGColor pcs
    where
      hasBGColor (Sq c _ _) = c `elem` [BLU,WSQ,BSQ,RED,PRP,GRN,DGRN,LBLU,MGRN,MRED,LGRN,LRED,SKYA,SKYB,SKYC,SKYD,FLRA,FLRB]
      hasBGColor (Pt c _)   = c `elem` [BLU,WSQ,BSQ,RED,PRP,GRN,DGRN,LBLU,MGRN,MRED,LGRN,LRED,SKYA,SKYB,SKYC,SKYD,FLRA,FLRB]
      hasFGColor (Sq c _ _) = c `elem` [WPC,BPC,BOR,RED,PRP,GRN,DGRN,LBLU,MGRN,MRED,LGRN,LRED,SKYA,SKYB,SKYC,SKYD,FLRA,FLRB]
      hasFGColor (Pt c _)   = c `elem` [WPC,BPC,BOR,RED,PRP,GRN,DGRN,LBLU,MGRN,MRED,LGRN,LRED,SKYA,SKYB,SKYC,SKYD,FLRA,FLRB]

spaceClose :: Window -> IO ()
spaceClose w = do k <- getKey w
                  if (k==' ' || k == '\x0') then closeWindow w
                                            else spaceClose w

-- Manipulating the drawings

flipV :: Picture -> Picture
flipV ((x,y), shapes) = ((x,y), Set.map f shapes)
    where
      f (Sq c p1 p2)   = Sq c (g p1) (g p2)
      f (Pt c (x1,y1)) = Pt c (x-x1-1,y1)
      g (x1,y1) = (x-x1,y1)

flipH :: Picture -> Picture
flipH ((x,y), shapes) = ((x,y), Set.map f shapes)
    where
      f (Sq c p1 p2) = Sq c (g p1) (g p2)
      f (Pt c (x1,y1)) = Pt c (x1,y-y1-1)
      g (x1,y1) = (x1,y-y1)


flipDiag :: Picture -> Picture
flipDiag ((x,y), shapes) = ((y,x), Set.map f shapes)
    where
      f (Sq c p1 p2) = Sq c (g p1) (g p2)
      f (Pt c pt)    = Pt c (g pt)
      g (x1,y1) = (y1,x1)

rotateL :: Picture -> Picture
rotateL = flipDiag . flipV

rotateR :: Picture -> Picture
rotateR = flipDiag . flipH

invertPiece :: Picture -> Picture
invertPiece ((x,y), shapes) = ((x,y), Set.map f shapes)
    where
      f (Sq c p1 p2) = Sq (g c) p1 p2
      f (Pt c pt)    = Pt (g c) pt
      g WPC = BPC
      g BPC = WPC
      g c = c

invertSquare :: Picture -> Picture
invertSquare ((x,y), shapes) = ((x,y), Set.map f shapes)
    where
      f (Sq c p1 p2) = Sq (g c) p1 p2
      f (Pt c pt)    = Pt (g c) pt
      g WSQ = BSQ
      g BSQ = WSQ
      g c = c

invert :: Picture -> Picture
invert = invertPiece . invertSquare

beside :: Picture -> Picture -> Picture
beside ((x1,y1), shapes1) ((x2,y2), shapes2) = ((x1+x2, max y1 y2), Set.union shapes1 (Set.map f shapes2))
    where
      f (Sq c p1 p2) = Sq c (g p1) (g p2)
      f (Pt c pt)    = Pt c (g pt)
      g (x,y) = (x+x1,y)

above :: Picture -> Picture -> Picture
above ((x1,y1), shapes1) ((x2,y2), shapes2) = ((max x1 x2, y1+y2), Set.union shapes1 (Set.map f shapes2))
    where
      f (Sq c p1 p2) = Sq c (g p1) (g p2)
      f (Pt c pt)    = Pt c (g pt)
      g (x,y) = (x,y+y1)

superimpose :: Picture -> Picture -> Picture
superimpose ((x1,y1), shapes1) ((x2,y2), shapes2) = ((max x1 x2, max y1 y2), Set.union shapes2 shapes1)
   
repeatH :: Int -> Picture -> Picture
repeatH n pc | n <= 1    = pc
             | otherwise = beside pc (repeatH (n-1) pc)

repeatV :: Int -> Picture -> Picture
repeatV n pc | n <= 1    = pc
             | otherwise = above pc (repeatV (n-1) pc)

rasterToPiece :: [String] -> Picture
rasterToPiece strs = ((x,y), Set.fromList (f 0 strs))
    where
      x = (length (head strs))*1
      y = (length strs)*1
      f _ [] = []
      f j (str:strs) = g 0 j str ++ f (j+1) strs
      g _ _ [] = []
      g i j (c:cs) | c == '#' = Pt BOR (i,j) : (g (i+1) j cs)
                   | c == '.' = Pt WPC (i,j) : (g (i+1) j cs)
                   | c == '1' = Pt RED (i,j) : (g (i+1) j cs)
                   | c == '2' = Pt PRP (i,j) : (g (i+1) j cs)
                   | c == '3' = Pt GRN (i,j) : (g (i+1) j cs)
                   | c == '4' = Pt DGRN (i,j) : (g (i+1) j cs)
                   | c == '5' = Pt LBLU (i,j) : (g (i+1) j cs)
                   | c == '6' = Pt MGRN (i,j) : (g (i+1) j cs)
                   | c == '7' = Pt MRED (i,j) : (g (i+1) j cs)
                   | c == '8' = Pt MGRN (i,j) : (g (i+1) j cs)
                   | c == '9' = Pt MRED (i,j) : (g (i+1) j cs)

                   | c == 'A' = Pt SKYA (i,j) : (g (i+1) j cs)
                   | c == 'B' = Pt SKYB (i,j) : (g (i+1) j cs)
                   | c == 'C' = Pt SKYC (i,j) : (g (i+1) j cs)
                   | c == 'D' = Pt SKYD (i,j) : (g (i+1) j cs)

                   | c == 'Z' = Pt FLRA (i,j) : (g (i+1) j cs)
                   | c == 'Y' = Pt FLRB (i,j) : (g (i+1) j cs)

                   | otherwise = (g (i+1) j cs)

-- The pieces themselves

whiteSquare, greySquare, blackSquare, clearSquare :: Picture
whiteSquare = ((58,58), Set.fromList [Sq WSQ (0,0) (58,58)])
greySquare  = ((58,58), Set.fromList [Sq BSQ (0,0) (58,58)])
blackSquare = greySquare
clearSquare = ((58,58), Set.empty)
bishop  = rasterToPiece bishopRaster
king    = rasterToPiece kingRaster
knight  = rasterToPiece knightRaster
pawn    = rasterToPiece pawnRaster
queen   = rasterToPiece queenRaster
rook    = rasterToPiece rookRaster
lambda  = rasterToPiece lambdaRaster
luke  = rasterToPiece lukeRaster



bishopRaster = ([
 "............#DDDDDDDDD#........",
 ".....#....##DDDDDD###########..",
 "##....####DDDDDDD#...........#.",
 "DD####DDDDDDDDDDD#.....##....##",
 "DDDDDDD#DDDDDDDDD#....#55#...#D",
 "DDDDDD#.#DDDDDDDD#....#55#...#D",
 "DDDDDD#.#DDDDDDDD#.....##....#D",
 "DDDDDD#.#DDDDDDDD#...........#D",
 "##DDDD#.#DDDDDDDD#...........#D",
 "#7#############################",
 "D#711111111111111111111777722##",
 "CCCC##71111111CCCCCCC777722##CC",
 "BBBCCCC#777CCCCBBBBBCC2222CCCCB",
 "55BBBBBCCCCBBBB5555BBBCCCCBBBB5",
 "5555555BBBB55555BBB555BBBB55555",
 "ABBBB5555555BBBBAAABBB555555BBB",
 " AAAABB555BBAAAA   AAABB55BBAAA",
 "     AABBBAA          AABBAA   ",
 "       AAA              AA     ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               "]
              )


kingRaster = ([
 "                               ",
 "                               ",
 "                               ",
 "AAA      AAAA             AAAAA",
 "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
 "ZZAAAAAZZZZZZZAAAAAAAAAZZZZZZZZ",
 "ZZZZAZZZZZZZZZZZZAAAZZZZZZZZYZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZYZZZZZZZZZZYZZZZZZZYZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZYZZ",
 "ZZZZZYZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZYZZZZZZZZZZYZZZZZZZ",
 "ZYZZZZZZZZZZZZZZZZZYZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZYZZZZZZZYZZZZZZZZZZZZYZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZYZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZYZZZZZ",
 "ZYZZZZZZZZZZYZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZYZZZZZZZZZZZZZZZZZZZZYZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZYZZZZZZZZZZZZZZZZZZZYZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZYZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZYZZZZZZZZZZZZZZZZZZZZ"]
              )

knightRaster = ([
 "ZZZZZZZZZZZZZZZZZZYZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZYZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZYZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZYZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZYZZZZZZZZZZZZZZZZZZZZZYZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZYZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZYZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZYZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZYZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZYZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZYZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZYZZZZZZZZZZZZZZZZZZZ",
 "ZYZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZYZ"]
                  )

lambdaRaster = ([
 "ZZZZZZZZZZZZZZZZZZYZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZYZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZYZZZZZ",
 "ZZZZZZZZZ72ZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZ71112ZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZ711112ZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZ711112ZZZZZZZZZZZZZZZZ",
 "ZZZZZYZZZZ711112ZZZZZZZZZZZYZZZ",
 "ZZZZZZZZZZZ711112ZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZ711112ZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZ71111112ZZZZZZZZZZZZ",
 "ZZZZZZZZZZ7111111112ZZZZZZZZZZZ",
 "ZZZZZZZZZ711111111112YZZZZZZZZZ",
 "ZZZZZZZZ711112ZZ111112ZZZZZZZZZ",
 "ZZZZYZZ711112ZZZZ111112ZZZZZZZZ",
 "ZZZZZZ711112ZZYZZZ111112ZZZZZZZ",
 "ZZZZZ711112ZZZZZZZZ111112ZZZZZZ",
 "ZZZYZ11112ZZZZZZZZZZ11112ZZZZZZ",
 "ZZZZZZ112ZZZZZZZZZZZZ112ZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZYZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZYZZZZZZZZZZZZZZZZZZZ",
 "ZYZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZYZ"]
                  )

lukeRaster = ([
 "ZZZZZZZZZZZZZZZZZZYZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZYZZZZZZZZZZZZZZZZZZZZZZZZ",
 "Z1ZZZ1ZZZ1Z1Z1ZZ1111ZZZZZYZZZZZ",
 "Z1ZZZ1ZZZ1Z1Z1ZZ1ZZZZZZZZZZZZZZ",
 "Z1ZZZ1ZZZ1Z11ZZZ111ZZZZZZZZZZZZ",
 "Z1ZZZ1ZZZ1Z1Z1ZZ1ZZZZZZZZZZZZZZ",
 "Z1111Z111ZZ1ZZ1Z1111ZZZZZZZZZZZ",
 "ZZZZZYZZZZZZZZZZZZZZZZZZZZZYZZZ",
 "Z111ZZ111ZZZ11ZZ1ZZZZZ1Z11ZZZ1Z",
 "Z1ZZ1Z1ZZ1Z1ZZ1Z1ZZZZZ1Z11ZZZ1Z",
 "Z111ZZ1ZZ1Z1YZ1ZZ1ZZZ1ZZ1Z1ZZ1Z",
 "Z1ZZ1Z111ZZ1ZZ1ZZ1Z1Z1ZZ1ZZ1Z1Z",
 "Z111ZZ1ZZ1ZZ11ZZZZ1Z1YZZ1ZZZ11Z",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZYZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZ111ZZZ111YZZ111ZZZ1111ZZZZ",
 "ZZZZ1ZZZ1Z1ZZZ1Z1ZZZ1Z1ZZZZ1ZZZ",
 "ZZZYZZZZ1Z1ZZZ1Z1ZZZ1Z111111ZZZ",
 "ZZZZZZZ1ZZ1ZZZ1Z1ZZZ1Z11ZZ11ZZZ",
 "ZZZZZZ1ZZZ1ZZZ1Z1ZZZ1Z1YZZZ1ZZZ",
 "ZZZZ11111ZZ111ZZZ111ZZZ1111ZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZYZZZZZZZZZZZZZZZZZZZ",
 "ZYZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ",
 "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZYZ"]
                  )


queenRaster = ([
 "                               ",
 "                               ",
 "    55     5               5   ",
 "   5. 5           6            ",
 "   5  5         864          86",
 "    55        8364          836",
 "            88364      5   836 ",
 "5         88333338        8336 ",
 "        883333333388     83336 ",
 "      883.333333333388  833336 ",
 "    8333...333636363336633336  ",
 "   83333.#.333636363664463336  ",
 "  8666636..63336363644  463336 ",
 "   444466663333336644    46336 ",
 "       444666666644   55  4636 ",
 "          4444464    5. 5  466 ",
 "    5          44    5  5   466",
 "                 4    55     46",
 "       5                       ",
 "99           55    5         99",
 "11999       5. 5          99911",
 "1111199     5  5  999   9911111",
 "111..119     55    279 91117111",
 "111.#1119           21711111717",
 "111.#7772           17277117711",
 "11177222      5    772 22777111",
 "77722             222    222777",
 "222     5                   222",
 "                               "]
              )

rookRaster = ([
 "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
 "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
 "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB",
 "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB",
 "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB",
 "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB",
 "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
 "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
 "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
 "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
 "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
 "CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDD####DDD####DDDDDDDDDDD",
 "DDDDDDDD#....###....#DDDDDDDDDD",
 "DDDDDDD#..........#..#DDDDDDDDD",
 "DDDDDDDD#.....####....#DDDDDDDD",
 "DDDDDDDDD#####DDDD####DDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "##DDDDDDDDDDDDDDDDDDDDDD###DD##",
 "..##DD###DDDDDDDDDDDD###...##..",
 "....##...#DDDDDDDDDD#........#.",
 "......#...#DDDDDDDDD#..........",
 "...........#DDDDDDDDD#........."]
              )


pawnRaster = ([
 "............#DDDDDDDDD#......#.",
 ".....#....##DDDDDDDDDD#......#.",
 "##....####DDDDDDDDDDD#......#..",
 "DD####DDDDDDDDDDDDDDDD#..###...",
 "DDDDDDDDDDDDDDDDDDDDDDD##DDD###",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
 "CCCCDDDDDDDDDDCCCCCCCDDDDDDDDCC",
 "BBBCCCCDDDDCCCCBBBBBCCDDDDCCCCB",
 "55BBBBBCCCCBBBB5555BBBCCCCBBBB5",
 "5555555BBBB55555BBB555BBBB55555",
 "ABBBB5555555BBBBAAABBB555555BBB",
 " AAAABB555BBAAAA   AAABB55BBAAA",
 "     AABBBAA          AABBAA   ",
 "       AAA              AA     ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               ",
 "                               "]
              )


